/*
 * tcpstream - Firestorm TCP state tracking implementation.
 * Copyright (c) 2002 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL v2
 *
 * TODO:
 *  o Configure the timeouts
 *  o Implement timeouts for SYN2/FIN states
 *  o Implement TIME_WAIT states
 *  o Re-order out-of-order transmissions
 *  o Intelligent stream reassembly
 *  o Support sack
 *  o Connection re-sync
 *  o Connection pickup ?
 *
 * Should CORRECTLY cope with:
 *  o Re-transmissions
 *  o Bad TCP or IP checksum evasion
 *  o Broadcast/Multicast packets
 *  o PAWS (rfc1323) evasion
*/

#ifndef DONT_INLINE
#define __INLINE__ static inline
#else
#define __INLINE__ static
#endif

#ifndef STATE_DEBUG
#define dmesg(x...)
#else
#define dmesg mesg
static char *state_str[]={
	"CLOSED",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TCP_TIME_WAIT",
	"TCP_CLOSE",
	"TCP_CLOSE_WAIT",
	"TCP_LAST_ACK",
	"TCP_LISTEN",
	"TCP_CLOSING"
};
#endif

#include <cleanup.h>
#include "tcpip.h"

#define TCP_HZ 100
#define TCP_PAWS_24DAYS (60 * 60 * 24 * 24)
#define TCP_PAWS_MSL 60
#define TCP_PAWS_WINDOW 60

#define TCP_TMO_SYN1 (90*TCP_HZ)

/* Configuration */
unsigned int tcp_minttl=0;
unsigned int tcp_numstreams=2048;
unsigned int tcp_numflows=1024;
unsigned int tcp_reassemble=0;
unsigned int tcp_stateful=0;

struct arg tcpstream_args[]={
	{"minttl", ARGTYPE_PUINT, NULL, {vp_uint:&tcp_minttl}},
	{"num_streams", ARGTYPE_PBYTES, NULL, {vp_bytes:&tcp_numstreams}},
	{"num_flows", ARGTYPE_PBYTES, NULL, {vp_bytes:&tcp_numflows}},
	{"reassemble", ARGTYPE_PBOOL, NULL, {vp_bool:&tcp_reassemble}},
	{NULL, ARGTYPE_NOP, NULL}
};

/* Token bucket for TCP state violations */
struct tokenbucket tcpstream_tb={
	.cost=(RATE_SEC/1), /* 1 alert per second */
	.burst=(RATE_SEC/1)*10, /* a burst of 10 alerts per second */
};

/* Token bucket for ICMP trying to fiddle with TCP state */
struct tokenbucket icmperr_tb={
	.cost=(RATE_SEC/1), /* 1 alert per second */
	.burst=(RATE_SEC/1)*10, /* a burst of 10 alerts per second */
};

/* Alert generator */
struct generator tcpstream_gen=init_generator("tcpstream", &tcpstream_tb);
struct generator icmperr_gen=init_generator("tcp.icmperr", &icmperr_tb);

/* Alerts */
struct alert alert_tcp1=init_alert("In-window SYN", 1, 0, 5);
struct alert alert_tcp2=init_alert("Data sent on closed stream", 2, 0, 5);
struct alert alert_tcp3=init_alert("PAWS discard", 3, 0, 5);
struct alert alert_icmp1=init_alert("ICMP error from interloper", 2, 0, 5);
struct alert alert_icmp2=init_alert("ICMP error on established connection", 2, 0, 5);

/* 1. Hash table for lookups -- O(n) in pathological cases, can be O(1) */
unsigned int tcp_hashsz=512;
struct tcp_session **tcp_hash=NULL;

/* 2. Slab cache for allocation -- O(1) */
union tcp_union *all_sessions=NULL;
union tcp_union *tcp_next=NULL;

/* 3. LRU list to evict oldest items -- O(1) */
struct tcp_lru lru={
	.next=(struct tcp_session *)&lru,
	.prev=(struct tcp_session *)&lru,
};

/* 4. Timeout list to evict entries on timeout */
struct tcp_tmo syn1={
	.next=(struct tcp_session *)&syn1,
	.prev=(struct tcp_session *)&syn1,
};

/* O(1) allocator for flows */
size_t flow_len=0;
void *flow_cache=NULL;
void *flow_next=NULL;

/* Statistics counters */
unsigned int tcp_packets=0;
unsigned int tcp_state_errs=0;
unsigned int tcp_broadcast=0;
unsigned int tcp_lowttl=0;
unsigned int tcp_timeouts=0;
unsigned int max_concurrent=0;
unsigned int num_active=0;
unsigned int max_flows=0;
unsigned int num_flows=0;

/* Wrap-safe seq/ack calculations */
__INLINE__ int before(u_int32_t s1, u_int32_t s2) {
	return (int32_t)(s1-s2) < 0;
}
__INLINE__ int after(u_int32_t s1, u_int32_t s2) {
	return (int32_t)(s2-s1) < 0;
}
__INLINE__ int between(u_int32_t s1, u_int32_t s2, u_int32_t s3) {
	return s3 - s2 >= s1 - s2; /* is s2<=s1<=s3 ? */
}
__INLINE__ u_int32_t tcp_receive_window(struct tcp_stream *tp)
{
	int32_t win=tp->rcv_wup+tp->rcv_wnd-tp->rcv_nxt;
	if ( win<0 )
		win=0;
	return (u_int32_t)win;
}
__INLINE__ int tcp_sequence(struct tcp_stream *tp, u_int32_t seq, u_int32_t end_seq)
{
	return !before(end_seq, tp->rcv_wup) &&
		!after(seq, tp->rcv_nxt+tcp_receive_window(tp));
}

/* Convert 64bit timestamp to 32bit jiffies */
__INLINE__ unsigned int tcp_jiffies(struct packet *p)
{
	unsigned int ret;
	ret=p->time.tv_sec * TCP_HZ;
	ret+=p->time.tv_usec / (1000000UL/TCP_HZ);
	return ret;
}

/* TMO: Add a TCP session to the head of a timeout list */
__INLINE__ void tcp_tmo_add(struct tcp_tmo *l, struct tcp_session *s)
{
	s->tmo_next=l->next;
	s->tmo_prev=(struct tcp_session *)l;
	l->next->tmo_prev=s;
	l->next=s;
}

/* TMO: Remove a TCP session from a timeout list */
__INLINE__ void tcp_tmo_del(struct tcp_session *s)
{
	if ( s->expire==0 )
		return;

	s->tmo_prev->tmo_next=s->tmo_next;
	s->tmo_next->tmo_prev=s->tmo_prev;
	s->tmo_prev=NULL;
	s->tmo_next=NULL;
	s->expire=0;
}

/* TMO: Check timeouts */
__INLINE__ void tcp_tmo_check(struct packet *pkt)
{
	unsigned int now=tcp_jiffies(pkt);

	while ( syn1.prev!=(struct tcp_session *)&syn1 ) {
		if ( (int)syn1.prev->expire - (int)now > 0 )
			return;
		tcp_free(syn1.prev);
		tcp_timeouts++;
	}
}

/* LRU: Add a TCP session to the head of an LRU list */
__INLINE__ void tcp_lru_add(struct tcp_lru *l, struct tcp_session *s)
{
	s->next=l->next;
	s->prev=(struct tcp_session *)l;
	l->next->prev=s;
	l->next=s;
}

/* LRU: Remove a TCP session from an LRU */
__INLINE__ void tcp_lru_del(struct tcp_session *s)
{
	s->prev->next=s->next;
	s->next->prev=s->prev;
}

/* LRU: Move a TCP session to the front of an LRU */
__INLINE__ void tcp_lru_mtf(struct tcp_lru *l, struct tcp_session *s)
{
	tcp_lru_del(s);
	tcp_lru_add(l,s);
}

/* HASH: Unlink a session from the session hash */
__INLINE__ void tcp_hash_unlink(struct tcp_session *s)
{
	if ( s->hash_next )
		s->hash_next->hash_pprev=s->hash_pprev;
	*s->hash_pprev=s->hash_next;
}

/* HASH: Link a session in to the TCP session hash */
__INLINE__ void tcp_hash_link(struct tcp_session *s)
{
	if ( (s->hash_next=tcp_hash[s->bucket]) ) {
		s->hash_next->hash_pprev=&s->hash_next;
	}
	tcp_hash[s->bucket]=s;
	s->hash_pprev=&tcp_hash[s->bucket];
}

/* HASH: Move to front of hash collision chain */
__INLINE__ void tcp_hash_mtf(struct tcp_session *s)
{
	tcp_hash_unlink(s);
	tcp_hash_link(s);
}

/* Flow cache allocation, slab-cache stylee */
/* TODO: strategy for when maximum is reached, evict LRU? */
__INLINE__ void *tcp_flow_alloc(void)
{
	void *ret=flow_next;
	if ( !ret ) return ret;
	flow_next=*(void **)flow_next;
	if ( ++num_flows > max_flows )
		max_flows=num_flows;
	return ret;
}

__INLINE__ void tcp_flow_free(void *p)
{
	*(void **)p=flow_next;
	flow_next=p;
	num_flows--;
}

/* Parse TCP options just for timestamps */
static int tcp_fast_options(struct pkt_tcphdr *t, u_int32_t *tsval)
{
	char *tmp, *end;

	/* Return if we don't have any */
	if ( t->doff<<2 <= sizeof(struct pkt_tcphdr)) return 0;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=(t->doff<<2);

	while ( tmp<end ) {
		if ( *tmp == TCPOPT_EOL || *tmp == TCPOPT_NOP ) {
			tmp++;
			continue;
		}

		if ( !(tmp+1 < end) ) break;

		switch(*tmp) {
		case TCPOPT_TIMESTAMP:
			if ( !(tmp+10 < end) ) break;
			*tsval=ntohl(*((u_int32_t *)(tmp+2)));
			return 1;
		}

		tmp+=*(tmp+1);
	}

	return 0;
}

/* This will parse TCP options for SYN packets */
static void tcp_syn_options(struct tcp_stream *s, struct pkt_tcphdr *t, u_int32_t sec)
{
	char *tmp, *end;

	/* Return if we don't have any */
	if ( t->doff<<2 <= sizeof(struct pkt_tcphdr)) return;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=(t->doff<<2);

	while ( tmp<end ) {
		if ( *tmp == TCPOPT_EOL || *tmp == TCPOPT_NOP ) {
			tmp++;
			continue;
		}

		if ( !(tmp+1 < end) ) break;

		switch(*tmp) {
		case TCPOPT_SACK_PERMITTED:
			s->flags|=TF_SACK_OK;
			break;
		case TCPOPT_TIMESTAMP:
			s->flags|=TF_TSTAMP_OK;

			/* Only check the bit we want */
			if ( !(tmp+10 < end) ) break;

			s->ts_recent=ntohl(*((u_int32_t *)(tmp+2)));
			s->ts_recent_stamp=sec;

			break;
		case TCPOPT_WSCALE:
			if ( !(tmp+2 < end) ) break;
			s->flags|=TF_WSCALE_OK;

			/* rfc1323: must log error and limit to 14 */
			if ( (s->scale=*(tmp+2)) > 14 ) s->scale=14;
			break;
		}

		tmp+=*(tmp+1);
	}
}

/* Perform a state transition */
__INLINE__ void transition(struct tcp_session *t, int c, int s)
{
#ifdef STATE_DEBUG
	char cip[32],sip[32];
	struct in_addr ca, sa;

	ca.s_addr=t->c_addr;
	sa.s_addr=t->s_addr;
	strncpy(cip, inet_ntoa(ca), sizeof(cip));
	strncpy(sip, inet_ntoa(sa), sizeof(sip));

	dmesg(M_DEBUG, "tcpstream: %s:%u(%s) -> %s:%u(%s)",
		cip, ntohs(t->c_port), state_str[c],
		sip, ntohs(t->s_port), state_str[s]);
#endif
	t->client.state=c;
	t->server.state=s;
}

/* Hash function. Hashes to the same value even
 * when source and destinations are inverted */
__INLINE__ int tcp_hashfn(u_int32_t saddr, u_int32_t daddr,
	u_int16_t sport, u_int16_t dport)
{
	int h = ((saddr ^ sport) ^ (daddr ^ dport));
	h ^= h>>16;
	h ^= h>>8;
	return h % tcp_hashsz;
}

/* Finds a protocol given a port number */
__INLINE__ struct proto *tcp_find_proto(u_int16_t port)
{
	struct proto_child *pc;

	/* IDs are server ports */
	for(pc=tcp_p.children; pc; pc=pc->next)
	{
		if ( port == pc->id )
			return pc->proto;
	}

	return NULL;
}

/* Return a TCP session descriptor to the cache */
void tcp_free(struct tcp_session *s)
{
	if ( !s ) {
		mesg(M_DEBUG,"tcpstream: severe insanity");
		return;
	}

	/* get the debug message */
	transition(s, 0, 0);

	/* HASH: unhash */
	tcp_hash_unlink(s);

	/* LRU: remove */
	tcp_lru_del(s);

	/* TMO: remove */
	tcp_tmo_del(s);

	if ( s->flow )
		tcp_flow_free(s->flow);

	/* CACHE: free */
	((union tcp_union *)s)->next=tcp_next;
	tcp_next=(union tcp_union *)s;

	num_active--;
}

/* Add a new session, never fails */
static struct tcp_session *tcp_new(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph, struct packet *p)
{
	struct tcp_session *ret;

	/* If we have run out of connections, then
	 * just free the oldest one in the LRU */
	if ( !tcp_next ) {
		tcp_free(lru.prev);
		tcp_timeouts++;
	}

	/* CACHE: alloc */
	ret=&tcp_next->s;
	tcp_next=tcp_next->next;

	/* Keep track of max number of concurrent sessions
	 * seen so far */
	if ( ++num_active > max_concurrent )
	       max_concurrent=num_active;

	/* Setup the key values */
	ret->c_addr=iph->saddr;
	ret->s_addr=iph->daddr;
	ret->c_port=tcph->sport;
	ret->s_port=tcph->dport;
	ret->bucket=tcp_hashfn(ret->c_addr, ret->s_addr,
		ret->c_port, ret->s_port);

	/* Setup initial state */
	transition(ret, TCP_SYN_SENT, 0);
	ret->server.isn=ntohl(tcph->seq);
	ret->client.snd_una=ret->server.isn+1;
	ret->client.snd_nxt=ret->client.snd_una+1;
	ret->client.rcv_wnd=ntohs(tcph->win);

	ret->server.rcv_nxt=ret->client.snd_una;
	ret->server.rcv_wup=ret->client.snd_una;

	/* server sees clients initial options */
	tcp_syn_options(&ret->server, tcph, p->time.tv_sec);

	if ( (ret->proto=tcp_find_proto(ret->s_port)) ) {
		ret->flow=tcp_flow_alloc();
	}else{
		/* TODO: auto-detect protocol ;) */
		ret->flow=NULL;
	}

	/* Set the timeout */
	ret->expire=tcp_jiffies(p)+TCP_TMO_SYN1;
	tcp_tmo_add(&syn1, ret);

	tcp_lru_add(&lru, ret);
	tcp_hash_link(ret);

	return ret;
}

/* Find a TCP session given a packet */
static struct tcp_session *tcp_find(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph,
	int *to_server)
{
	struct tcp_session *s;
	int idx=tcp_hashfn(iph->saddr, iph->daddr,
		tcph->sport, tcph->dport);

	for(s=tcp_hash[idx]; s; s=s->hash_next) {
		if (	s->s_addr==iph->saddr &&
			s->c_addr==iph->daddr &&
			s->s_port==tcph->sport &&
			s->c_port==tcph->dport ) {
			*to_server=0;
			return s;
		}
		if (	s->c_addr==iph->saddr &&
			s->s_addr==iph->daddr &&
			s->c_port==tcph->sport &&
			s->s_port==tcph->dport ) {
			*to_server=1;
			return s;
		}
	}

	return NULL;
}

/* rfc793: Actions to perform when recieving an ACK in
 * an established state */
__INLINE__ void tcp_established(struct tcp_stream *snd,
	struct tcp_stream *rcv,
	u_int32_t seq,
	u_int32_t ack,
	u_int32_t win)
{
	if ( after(ack, rcv->snd_una) && !before(ack, rcv->snd_nxt) ) {
		rcv->snd_una=ack;
		rcv->rcv_nxt=seq+1;
		snd->rcv_wup=ack;
		snd->rcv_wnd=win;
	}

	snd->snd_una=seq;
	snd->snd_nxt=seq+1;
}

/* A segment contained valid data so attempt
 * stream reassembly -- this is not finished yet */
static void tcpstream_data(
	struct tcp_session *s,
	struct tcp_stream *rcv,
	struct pkt_tcphdr *tcph,
	struct tcpseg *tp)
{
	/* char *ptr=(char *)tcph + (tcph->doff<<2); */

	if ( tp->seq==rcv->rcv_nxt ) {
		if (tcp_receive_window(rcv) == 0) {
			/* XXX: Alert here? Sending beyond end of window.
			 * Some buggy stacks do this a lot... */
			return;
		}

		dmesg(M_DEBUG, "tcpstream: In order: %u->%u",
			tp->seq-rcv->isn,
			tp->seq_end-rcv->isn);

		/* XXX: Can only deliver data if rcv is in ESTABLISHED,
		 * FIN_WAIT1 or FIN_WAIT2 -- hrmmm */
		/* TODO: Intelligent reassembly
		 *  1. Application layer plugin recives shit
		 *  2. Application layer plugin signals us to start buffering
		 *  3. Here we aggregate each new packet that comes in
		 *  4. Free buffer if size>max || app signals OK
		 */
		rcv->rcv_nxt=tp->seq_end;
	}else{
		/* A retransmit */
		if ( !after(tp->seq_end,rcv->rcv_nxt) )
			return;

		/* TODO: Always need to re-order any out of order
		 * transmissions. Not sure what kind of mem policy
		 * we are going to need here...
		 */
		dmesg(M_DEBUG, "tcpstream: Out of order: %u->%u %u->%u",
			tp->seq-rcv->isn,
			tp->seq_end-rcv->isn,
			tp->seq, rcv->rcv_nxt);
	}
}

/* Process incoming TCP segments here */
static int tcpstream_tcpseg(struct packet *p, unsigned int i)
{
	struct pkt_iphdr *iph=p->layer[i].h.ip;
	struct pkt_tcphdr *tcph=p->layer[i+1].h.tcp;
	struct tcp_session *s;
	struct tcp_stream *snd, *rcv;
	int to_server;
	int ret=0;
	struct tcpseg tp;

	/* Host order header values */
	tp.ack=ntohl(tcph->ack);
	tp.seq=ntohl(tcph->seq);
	tp.win=ntohs(tcph->win);
	tp.len=ntohs(iph->tot_len) - (iph->ihl<<2) - (tcph->doff<<2);
	tp.seq_end=tp.seq+tp.len+tcph->flags.bits.syn+tcph->flags.bits.fin;
	tp.tsval=0;
	tp.saw_tstamp=0;

	/* Find the session this packet comes from */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		if ( (tcph->flags.bits.syn) &&
			!(tcph->flags.bits.ack) &&
			!(tcph->flags.bits.rst)) {

			/* SYN: part 1 of connection handshake */
			p->layer[i+1].session=tcp_new(iph, tcph, p);
			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_2SVR;
			return ret;
		}

		dmesg(M_DEBUG,"tcpstream: never heard of you");
		return ret;
	}else p->layer[i+1].session=s;

	/* figure out who is sender and who is reciever */
	if ( to_server ) {
		snd=&s->client;
		rcv=&s->server;
		p->layer[i+1].flags|=FLAG_TCP_2SVR;
	}else{
		snd=&s->server;
		rcv=&s->client;
		p->layer[i+1].flags&=(~FLAG_TCP_2SVR);
	}

	/* LRU: Move to front of LRU and hash collision chain */
	tcp_lru_mtf(&lru, s);
	tcp_hash_mtf(s);

	/* Deal with a SYN/ACK */
	if ( rcv->state==TCP_SYN_SENT ) {
		/* fist check the ack field */
		if ( tcph->flags.bits.ack ) {
		 	/* if SND.UNA =< SEG.ACK =< SND.NXT
			 * then ACK is acceptable */
			if ( !(between(tp.ack, rcv->snd_una, rcv->snd_nxt)) ) {
				return ret;
			}

			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
		}

		/* then check the rst flag */
		if ( tcph->flags.bits.rst ) {
			return 1;
		}

		if ( tcph->flags.bits.syn ) {
			/* update the advertised window */
			snd->rcv_wnd=tp.win;

			/* update sequencing information */
			snd->snd_una=tp.seq+1;
			snd->snd_nxt=snd->snd_una+1;

			rcv->isn=tp.seq;
			rcv->rcv_nxt=tp.seq_end;
			rcv->rcv_wup=tp.seq_end;
			if ( tcph->flags.bits.ack )
				rcv->snd_una=tp.ack;

			/* client now sees servers initial options */
			tcp_syn_options(&s->client, tcph, p->time.tv_sec);

			/* Check whether to use window scaling */
			if (!(rcv->flags&TF_WSCALE_OK && snd->flags&TF_WSCALE_OK)) {
				rcv->scale=0;
				snd->scale=0;
			}

			/* SYN|ACK: part 2 of connection handshake */
			rcv->state=TCP_SYN_RECV;
			snd->state=TCP_SYN_SENT;
			transition(s, s->client.state, s->server.state);
			tcp_tmo_del(s);
			return ret;
		}else goto no_checks;
	}

	/* First check the sequence number */
	if ( !tcp_sequence(rcv, tp.seq, tp.seq_end) ) {
		/* XXX: Don't alert here retransmits hit this */
		return ret;
	}

no_checks:

	/* rfc1323: H1. Apply PAWS checks first */
	if ( rcv->flags&TF_TSTAMP_OK &&
		(tp.saw_tstamp=tcp_fast_options(tcph, &tp.tsval)) ) {
		/* TODO: Fix PAWS check, its buggy as fuck */
		if ( (int32_t)(rcv->ts_recent - tp.tsval) > TCP_PAWS_WINDOW &&
			p->time.tv_sec<rcv->ts_recent_stamp + TCP_PAWS_24DAYS ) {
			alert(&tcpstream_gen, p, &alert_tcp3);
			return ret;
		}
	}

	/* Second, check the RST bit */
	if ( tcph->flags.bits.rst ) {
		p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
		dmesg(M_DEBUG,"TCP stream was reset");
		return 1;
	}

	/* rfc1323: PAWS: update ts_recent */
	if ( tp.saw_tstamp && !after(tp.seq,rcv->rcv_wup) ) {
		if((int32_t)(tp.tsval - rcv->ts_recent) >= 0 ||
		   p->time.tv_sec >= rcv->ts_recent_stamp + TCP_PAWS_24DAYS) {
			rcv->ts_recent=tp.tsval;
			rcv->ts_recent_stamp=p->time.tv_sec;
		}
	}

	/* Third, check security and precendece (pfft) */

	/* Fourth, check the SYN bit */
	if ( tcph->flags.bits.syn && !before(tp.seq,rcv->rcv_nxt) ) {
		alert(&tcpstream_gen, p, &alert_tcp1);
		return 1;
	}

	/* we now know that this packet is in-state */
	p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;

	/* Scale the window */
	tp.win<<=snd->scale;

	/* Fifth, check the ack field */
	if ( !tcph->flags.bits.ack ) goto no_ack;
	switch(rcv->state) {
	case TCP_SYN_SENT:
		/* ACK: part 3 of connection handshake */
		if ( rcv->snd_una <= tp.ack && tp.ack <= rcv->snd_nxt ) {
			p->layer[i+1].flags|=FLAG_TCP_CT_EST;
			tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
			transition(s, TCP_ESTABLISHED, TCP_ESTABLISHED);
		}
		break;
	case TCP_ESTABLISHED:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		rcv->rcv_nxt=tp.seq;
		break;
	case TCP_FIN_WAIT1:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		rcv->state=TCP_FIN_WAIT2;
		snd->state=TCP_LAST_ACK;
		transition(s, s->client.state, s->server.state);
	case TCP_FIN_WAIT2:
		break;
	case TCP_CLOSE_WAIT:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		break;
	case TCP_CLOSING:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_LAST_ACK:
		rcv->state=0;
		transition(s, s->client.state, s->server.state);
		break;
	default:
		break;
	}

no_ack:
	/* sixth check URG bit */
	if ( !tcph->flags.bits.urg ) goto no_urg;

	/* XXX: What does processing urgent/OOB data
	 * actually buy us in terms of intrusion detection
	 * capability? Need to look in to this... */

no_urg:
	/* seventh process the segment text */

	/* We might not actually need to bother... */
	if ( !tcp_reassemble ) goto no_data;
	if ( !tp.len || !s->proto ) goto no_data;

	/* There is data in the segment */
	switch(rcv->state) {
	case TCP_CLOSE_WAIT:
	case TCP_CLOSING:
	case TCP_LAST_ACK:
	case TCP_TIME_WAIT:
		alert(&tcpstream_gen, p, &alert_tcp2);
		goto no_data;
	}

	tcpstream_data(s, rcv, tcph, &tp);

no_data:
	/* eighth, check the FIN bit */
	if ( !tcph->flags.bits.fin ) goto no_fin;

	/* XXX: Fin implies PSH */
	rcv->rcv_nxt++;

	switch(rcv->state) {
	case TCP_SYN_RECV:
	case TCP_ESTABLISHED:
		rcv->state=TCP_CLOSE_WAIT;
		snd->state=TCP_FIN_WAIT1;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT1:
		/* if fin has been acked do time_wait else closing */
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT2:
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_CLOSE_WAIT:
	case TCP_CLOSING:
	case TCP_LAST_ACK:
	case TCP_TIME_WAIT:
		/* Remain in the same state */
		break;
	default:
		break;
	}

no_fin:
	/* XXX: need to implement time_wait timer */
	if ( (snd->state==TCP_TIME_WAIT && rcv->state==0) ||
		(rcv->state==TCP_TIME_WAIT && snd->state==0) ) {
		ret=1;
	}else if ( rcv->state==0 && snd->state==TCP_SYN_SENT ) {
		/* XXX: we reset the timeout on retransmissions */
		if ( (tcph->flags.bits.syn) &&
			!(tcph->flags.bits.ack) &&
			!(tcph->flags.bits.rst)) {
			tcp_tmo_del(s);
			s->expire=tcp_jiffies(p)+TCP_TMO_SYN1;
			tcp_tmo_add(&syn1, s);
		}
	}

	return ret;
}

/* Process an ICMP error message during connection */
void icmperr_process(struct packet *p, unsigned int l)
{
	struct pkt_iphdr *outer_iph;
	struct pkt_icmphdr *icmph;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	struct tcp_session *s;
	struct tcp_stream *snd,*rcv;
	int to_server;

	/* Get all the headers we need */
	if ( l+2 >= p->llen ) return;
	outer_iph=p->layer[l-1].h.ip;
	icmph=p->layer[l].h.icmp;
	iph=p->layer[l+1].h.ip;
	tcph=p->layer[l+2].h.tcp;

	/* Make sure the packet is cool */
	if ( iph->protocol!=6 ) return;
	if ( icmph->type != ICMP_DEST_UNREACH ) return;
	if ( icmph->code == ICMP_FRAG_NEEDED ) return;
	if ( icmph->code > NR_ICMP_UNREACH ) return;

	/* Check the IP packet came from the endpoint
	 * ( if necessary ) */
	if ( icmph->code==ICMP_PROT_UNREACH ||
		icmph->code==ICMP_PORT_UNREACH ) {
		if ( iph->saddr!=outer_iph->daddr ) {
			alert(&icmperr_gen, p, &alert_icmp1);
			return;
		}
	}

	/* Lookup our session */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		/* XXX: Alert here? Seems to happen a lot... */
		return;
	}

	/* Find out which side sent the offending packet */
	if ( iph->daddr==s->c_addr ) {
		snd=&s->server;
		rcv=&s->client;
	}else{
		snd=&s->client;
		rcv=&s->server;
	}

	/* Only reset opening connections */
	if ( snd->state!=TCP_SYN_SENT && snd->state!=TCP_SYN_RECV ) {
		alert(&icmperr_gen, p, &alert_icmp2);
		return;
	}

	/* TODO: Check seq/ack numbers */
	dmesg(M_DEBUG,"TCP reset by ICMP from %.8x", outer_iph->saddr);
	tcp_free(s);
}

/* Do some basic sanity checks to decide whether to even
 * bother processing a given segment or not. */
void tcpstream_process(struct packet *pkt, unsigned int i)
{
	struct pkt_iphdr *iph=pkt->layer[i].h.ip;
	struct layer *l=&pkt->layer[i+1];
	struct tcp_session *s;
	int ret=0;

	l->flags|=FLAG_TCP_TRACK;

	/* Ignore fragments */
	if ( iph->frag_off & ipfmask )
		goto nodecode;

	/* Ignore bad IP checksums - ooh naughty! */
	if ( !(pkt->layer[i].flags&FLAG_IP_CSUM) )
		goto nodecode;

	/* Ignore bad TCP checksums - bad boy! */
	if ( !(l->flags&FLAG_TCP_CSUM) )
		goto nodecode;

	/* Ignore broadcast and multicast packets */
	if ( (pkt->flags & (FP_MULTICAST|FP_BROADCAST)) ) {
		tcp_broadcast++;
		goto nodecode;
	}

	/* Ignore packets with TTLs lower than minttl */
	if ( iph->ttl < tcp_minttl ) {
		tcp_lowttl++;
		goto nodecode;
	}

	/* Check we aren't in a icmp error message */
	if ( i>0 && pkt->layer[i-1].proto==&icmp_p )
		goto nodecode;

	/* Actually do the state tracking */
	ret=tcpstream_tcpseg(pkt, i);

	/* Statistics counters */
	if ( !(l->flags & FLAG_TCP_STATE) )
		tcp_state_errs++;
	tcp_packets++;

	/* Have we actually been decoded? */
	if ( !(s=l->session) || !s->proto )
		goto nodecode;

	/* Dispatch to protocol handler */
	pkt->layer[pkt->llen].flags=0;
	pkt->layer[pkt->llen].session=NULL;
	pkt->layer[pkt->llen].proto=s->proto;
	if ( s->proto->sdecode(pkt)<0 && !ret ) {
		/* TODO: queue this packet for reassembly */
		mesg(M_DEBUG, "queue");
	}

	goto done;
nodecode:
	if ( pkt->layer[pkt->llen].h.raw<pkt->end )
		pkt->layer[pkt->llen++].proto=NULL;
	dispatch(pkt);
done:
	if (ret) tcp_free(l->session);
	tcp_tmo_check(pkt);
	return;
}

/* Called before firestorm shuts down */
void tcpstream_free(void)
{
	mesg(M_INFO,"tcpstream: max_concurrent=%u num_active=%u",
		max_concurrent, num_active);

	if ( tcp_reassemble ) {
		mesg(M_INFO,"tcpstream: max_flows=%u num_flows=%u",
			max_flows, num_flows);
	}

	mesg(M_INFO,"tcpstream: %u state errors out of %u packets",
	       tcp_state_errs, tcp_packets);
	mesg(M_INFO,"tcpstream: %u broadcasts, %u ttl evasions, %u timeouts",
		tcp_broadcast, tcp_lowttl, tcp_timeouts);

	if ( tcp_hash )
		free(tcp_hash);

	if ( all_sessions )
		free(all_sessions);
}

/* Initialise flow cache */
int tcpstream_flow_init(void)
{
	void *p,*n=NULL;
	int i;

	/* Pointless if not reassembling, or have
	 * no application layer protocols with flow
	 * capabilities... */
	if ( !tcp_reassemble || !flow_len )
		return 0;

	/* not likely */
	if ( flow_len < sizeof(void *) ) {
		flow_len=sizeof(void *);
	}

	/* flows are 1 to 1 with streams, even in worst case */
	if ( tcp_numflows > tcp_numstreams ) {
		tcp_numflows=tcp_numstreams;
		mesg(M_WARN, "tcpstream: num_flows>num_streams, duh");
	}

	if ( tcp_numflows < 256 ) {
		mesg(M_WARN, "tcpstream: you may be at risk with "
			"such a low num_flows value");
	}

	/* Allocate it */
	if ( !(flow_cache=calloc(tcp_numflows, flow_len)) ) {
		return -1;
	}

	/* Initialise it */
	for(i=0; i<(tcp_numflows-1); ) {
		p=(char *)flow_cache + (i*flow_len);
		n=(char *)flow_cache + (++i*flow_len);
		*(void **)p = n;
	}

	*(void **)n=NULL;
	flow_next=flow_cache;

	return 0;
}

/* Initialise based on user parameters */
int tcpstream_init(char *args)
{
	int i;

	if ( tcp_stateful ) {
		mesg(M_ERR,"tcpstream: Can't add tcpstream twice!");
		return 0;
	}

	if ( args && args_parse(tcpstream_args, args, NULL)<=0 ) {
		mesg(M_ERR,"tcpstream: parse error: %s", args);
		return 0;
	}

	if ( tcp_minttl > 255 ) {
		mesg(M_ERR,"tcpstream: minttl must be < 256");
		return 0;
	}

	if ( tcp_numstreams < 32 ) {
		mesg(M_WARN,"tcpstream: num_streams is stupidly "
			"low, setting to 512");
		tcp_numstreams=512;
	}else if ( tcp_numstreams < 512 ) {
		mesg(M_WARN,"tcpstream: num_streams is VERY low! "
			"are you sure?!");
	}

	/* Select a size for the hash table such that
	 * num_streams/hash_size is no larger than 128
	 * and such that the minimum size is 512 */
	for(tcp_hashsz=512;
		tcp_numstreams/tcp_hashsz>128;
		tcp_hashsz<<=1);

	/* Make sure its always odd */
	tcp_hashsz|=1;

	/* Initialise data structures */
	if ( !(all_sessions=calloc(tcp_numstreams,
		sizeof(struct tcp_session))) ) {
		goto oom;
	}
	if ( !(tcp_hash=calloc(tcp_hashsz,
		sizeof(*tcp_hash))) ) {
		free(all_sessions);
		goto oom;
	}

	/* Initialise the slab-style allocator */
	for(i=0; i<(tcp_numstreams-1); i++) {
		all_sessions[i].next=&all_sessions[i+1];
	}
	all_sessions[i].next=NULL;
	tcp_next=all_sessions;

	/* Initialise state info */
	if ( tcp_reassemble ) {
		struct proto_child *pc;

		for(pc=tcp_p.children; pc; pc=pc->next)
		{
			if ( pc->proto->state_len > flow_len )
				flow_len = (size_t)pc->proto->state_len;
		}

	}

	/* Initialise flow cache */
	if ( tcpstream_flow_init() )
		goto oom;

	/* Print out some statistics */
	mesg(M_INFO,"tcpstream: %u streams in %u buckets (%u KB)",
		tcp_numstreams, tcp_hashsz,
		(tcp_numstreams*sizeof(struct tcp_session))/1024);
	mesg(M_INFO,"tcpstream: TCP stream reassembly is %s: %u flows",
		tcp_reassemble ? "ENABLED" : "DISABLED",
		tcp_reassemble ? tcp_numflows : 0);

	tcp_stateful=1;

	return 1;
oom:
	mesg(M_ERR,"tcpstream: calloc(): %s", get_err());
	return 0;
}
